/*
 *
 * (C) COPYRIGHT 2014-2017 MStar Semiconductor, Inc. All rights reserved.
 *
 * This program is free software and is provided to you under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation, and any use by you of this program is subject to the terms
 * of such GNU licence.
 *
 * A copy of the licence is included with the program, and can also be obtained
 * from Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 */

#include <mali_kbase.h>
#include "mali_kbase_platform_mstar.h"
#include "mali_kbase_platform_mstar_m7821_reg.h"

#include <linux/delay.h>
#include "mstar/mstar_chip.h"

/* Platform functions */
void init_registers(void)
{
#define MIU_BASE_TO_MB(base)        ((base) >> 20)                   /* in MB */
#define DRAM_SIZE_REG_TO_MB(reg)    ((reg > 0) ? (1 << (reg)) : 0)   /* in MB */

    /* set GPU clock */
    SET_REG(gpu_pll, set_lo, (MHZ_TO_REG(MALI_MAX_FREQ)>> 0)&0xffff);
    SET_REG(gpu_pll, set_hi, (MHZ_TO_REG(MALI_MAX_FREQ)>>16)&0x00ff);
    SET_REG(gpu_pll, sld, 1);
    udelay(10);

    /* ECO: use DVDD CLK Output port: ARMPLL_VCO_DIV2 */
    /* set ARMPLL_EN_OCCCLK2=1 */
    SET_REG(gpu_pll, g3d_pll, 3);
    SET_REG(gpu_pll, set_armpll_post, 0);
    /* set ARMPLL_POST_DIVN=2' b01 */
    SET_REG(gpu_pll, set_armpll_post_divn, 0x41);
    udelay(1);

    /* set MIU addr */
    SET_REG(gpu, miu0_start, MIU_BASE_TO_MB(MSTAR_MIU0_BUS_BASE));
    SET_REG(gpu, miu0_end, MIU_BASE_TO_MB(MSTAR_MIU0_BUS_BASE) + DRAM_SIZE_REG_TO_MB(GET_REG(miu0, dram_size)));
#if defined(MSTAR_MIU1_BUS_BASE) && defined(ARM_MIU1_BASE_ADDR)
    SET_REG(gpu, miu1_start, MIU_BASE_TO_MB(MSTAR_MIU1_BUS_BASE));
    SET_REG(gpu, miu1_end, MIU_BASE_TO_MB(MSTAR_MIU1_BUS_BASE) + DRAM_SIZE_REG_TO_MB(GET_REG(miu1, dram_size)));
#else
    SET_REG(gpu, miu1_start, 0xffff);
    SET_REG(gpu, miu1_end, 0xffff);
#endif

    /* enable RIU */
#ifdef MSTAR_RIU_ENABLED
    SET_REG(gpu, riu, 1);
#endif

    /* disable MF decoder */
    SET_REG(gpu, mfdec, 1);

    /* disable gpu_always_power_on */
    SET_REG(gpu, always_on, 0);

    /* power on G3DPLL */
    SET_REG(gpu_pll, g3d_pll, 2);
    udelay(1);
}

void power_on(void)
{
    SET_REG(gpu, sram_sd_en, 0);    /* power on SRAM */
    SET_REG(clkgen1, ckg_gpu, 0);   /* enable GPU clock */

    /* reset: 0 -> 1 -> 0 */
    SET_REG(gpu, soft_reset, 0);
    SET_REG(gpu, soft_reset, 1);;
    udelay(1); /* delay for runtime suspend */
    SET_REG(gpu, soft_reset, 0);
    udelay(1);
}

void power_off(void)
{
    SET_REG(gpu, soft_reset, 1);    /* reset GPU */
    SET_REG(clkgen1, ckg_gpu, 1);   /* disable GPU clock */
    SET_REG(gpu, sram_sd_en, 1)     /* power off SRAM */
    udelay(1);
}

void soft_reset(void)
{
    /* This callback function is for GPU secure mode use.
     * If this chip need to support GPU secure mode, refer to m7332,
     * or chip spec to implement it.
     * keep it empty by default.
     */
}

/* Patch HW Bug: gpu software reset will cause FIFO points in disarray*/
void power_suspend(void)
{
    /* power off */
    /* don't reset GPU*/
    SET_REG(clkgen1, ckg_gpu, 1);   /* disable GPU clock */
    SET_REG(gpu, sram_sd_en, 1)     /* power off SRAM */
    udelay(1);
}

/* Patch HW Bug: gpu software reset will cause FIFO points in disarray*/
void power_resume(void)
{
    init_registers();

    /* power on */
    SET_REG(gpu, sram_sd_en, 0);    /* power on SRAM */
    SET_REG(clkgen1, ckg_gpu, 0);   /* enable GPU clock */
    /* don't reset GPU*/
    udelay(1);
}

#ifdef MSTAR_DISABLE_SHADER_CORES
int get_num_disabled_cores(void)
{
    if (GET_REG(gpu, disable_half_cores) == 1)
        return 2;
    return 0;
}
#endif

void get_dram_length(u32* miu0_length, u32* miu1_length, u32* miu2_length)
{
#define DRAM_SIZE_REG_TO_MB(reg)    ((reg > 0) ? (1 << (reg)) : 0)      /* in MB */

    *miu0_length = DRAM_SIZE_REG_TO_MB(GET_REG(miu0, dram_size));
#if defined(MSTAR_MIU1_BUS_BASE) && defined(ARM_MIU1_BASE_ADDR)
    *miu1_length = DRAM_SIZE_REG_TO_MB(GET_REG(miu1, dram_size));
#else
    *miu1_length = 0;
#endif
    *miu2_length = 0;

    return;
}
